@@ -9,58 +9,60 @@ const formTags = new Set([
     "datalist",
     "textarea",
 ]);
-
 const pTag = new Set(["p"]);
+const tableSectionTags = new Set(["thead", "tbody"]);
+const ddtTags = new Set(["dd", "dt"]);
+const rtpTags = new Set(["rt", "rp"]);
 
-const openImpliesClose: Record<string, Set<string>> = {
-    tr: new Set(["tr", "th", "td"]),
-    th: new Set(["th"]),
-    td: new Set(["thead", "th", "td"]),
-    body: new Set(["head", "link", "script"]),
-    li: new Set(["li"]),
-    p: pTag,
-    h1: pTag,
-    h2: pTag,
-    h3: pTag,
-    h4: pTag,
-    h5: pTag,
-    h6: pTag,
-    select: formTags,
-    input: formTags,
-    output: formTags,
-    button: formTags,
-    datalist: formTags,
-    textarea: formTags,
-    option: new Set(["option"]),
-    optgroup: new Set(["optgroup", "option"]),
-    dd: new Set(["dt", "dd"]),
-    dt: new Set(["dt", "dd"]),
-    address: pTag,
-    article: pTag,
-    aside: pTag,
-    blockquote: pTag,
-    details: pTag,
-    div: pTag,
-    dl: pTag,
-    fieldset: pTag,
-    figcaption: pTag,
-    figure: pTag,
-    footer: pTag,
-    form: pTag,
-    header: pTag,
-    hr: pTag,
-    main: pTag,
-    nav: pTag,
-    ol: pTag,
-    pre: pTag,
-    section: pTag,
-    table: pTag,
-    ul: pTag,
-    rt: new Set(["rt", "rp"]),
-    rp: new Set(["rt", "rp"]),
-    tbody: new Set(["thead", "tbody"]),
-    tfoot: new Set(["thead", "tbody"]),
-};
+const openImpliesClose = new Map<string, Set<string>>([
+    ["tr", new Set(["tr", "th", "td"])],
+    ["th", new Set(["th"])],
+    ["td", new Set(["thead", "th", "td"])],
+    ["body", new Set(["head", "link", "script"])],
+    ["li", new Set(["li"])],
+    ["p", pTag],
+    ["h1", pTag],
+    ["h2", pTag],
+    ["h3", pTag],
+    ["h4", pTag],
+    ["h5", pTag],
+    ["h6", pTag],
+    ["select", formTags],
+    ["input", formTags],
+    ["output", formTags],
+    ["button", formTags],
+    ["datalist", formTags],
+    ["textarea", formTags],
+    ["option", new Set(["option"])],
+    ["optgroup", new Set(["optgroup", "option"])],
+    ["dd", ddtTags],
+    ["dt", ddtTags],
+    ["address", pTag],
+    ["article", pTag],
+    ["aside", pTag],
+    ["blockquote", pTag],
+    ["details", pTag],
+    ["div", pTag],
+    ["dl", pTag],
+    ["fieldset", pTag],
+    ["figcaption", pTag],
+    ["figure", pTag],
+    ["footer", pTag],
+    ["form", pTag],
+    ["header", pTag],
+    ["hr", pTag],
+    ["main", pTag],
+    ["nav", pTag],
+    ["ol", pTag],
+    ["pre", pTag],
+    ["section", pTag],
+    ["table", pTag],
+    ["ul", pTag],
+    ["rt", rtpTags],
+    ["rp", rtpTags],
+    ["tbody", tableSectionTags],
+    ["tfoot", tableSectionTags],
+]);
 
 const voidElements = new Set([
     "area",
@@ -93,7 +95,7 @@ const htmlIntegrationElements = new Set([
     "ms",
     "mtext",
     "annotation-xml",
-    "foreignObject",
+    "foreignobject",
     "desc",
     "title",
 ]);
@@ -165,7 +167,7 @@ export interface Handler {
      */
     onend(): void;
     onerror(error: Error): void;
-    onclosetag(name: string): void;
+    onclosetag(name: string, isImplied: boolean): void;
     onopentagname(name: string): void;
     /**
      *
@@ -178,7 +180,11 @@ export interface Handler {
         value: string,
         quote?: string | undefined | null
     ): void;
-    onopentag(name: string, attribs: { [s: string]: string }): void;
+    onopentag(
+        name: string,
+        attribs: { [s: string]: string },
+        isImplied: boolean
+    ): void;
     ontext(data: string): void;
     oncomment(data: string): void;
     oncdatastart(): void;
@@ -193,7 +199,12 @@ export class Parser {
     /** The start index of the last event. */
     public startIndex = 0;
     /** The end index of the last event. */
-    public endIndex: number | null = null;
+    public endIndex = 0;
+    /**
+     * Store the start index of the current open tag,
+     * so we can update the start index for attributes.
+     */
+    private openTagStart = 0;
 
     private tagname = "";
     private attribname = "";
@@ -202,13 +213,14 @@ export class Parser {
     private stack: string[] = [];
     private readonly foreignContext: boolean[] = [];
     private readonly cbs: Partial<Handler>;
-    private readonly options: ParserOptions;
     private readonly lowerCaseTagNames: boolean;
     private readonly lowerCaseAttributeNames: boolean;
     private readonly tokenizer: Tokenizer;
 
-    constructor(cbs: Partial<Handler> | null, options: ParserOptions = {}) {
-        this.options = options;
+    constructor(
+        cbs?: Partial<Handler> | null,
+        private readonly options: ParserOptions = {}
+    ) {
         this.cbs = cbs ?? {};
         this.lowerCaseTagNames = options.lowerCaseTags ?? !options.xmlMode;
         this.lowerCaseAttributeNames =
@@ -220,46 +232,48 @@ export class Parser {
         this.cbs.onparserinit?.(this);
     }
 
-    private updatePosition(initialOffset: number) {
-        if (this.endIndex === null) {
-            if (this.tokenizer.sectionStart <= initialOffset) {
-                this.startIndex = 0;
-            } else {
-                this.startIndex = this.tokenizer.sectionStart - initialOffset;
-            }
-        } else {
-            this.startIndex = this.endIndex + 1;
-        }
-        this.endIndex = this.tokenizer.getAbsoluteIndex();
-    }
-
     // Tokenizer event handlers
+
+    /** @internal */
     ontext(data: string): void {
-        this.updatePosition(1);
-        (this.endIndex as number)--;
+        const idx = this.tokenizer.getAbsoluteIndex();
+        this.endIndex = idx - 1;
         this.cbs.ontext?.(data);
+        this.startIndex = idx;
     }
 
+    protected isVoidElement(name: string): boolean {
+        return !this.options.xmlMode && voidElements.has(name);
+    }
+
+    /** @internal */
     onopentagname(name: string): void {
+        this.endIndex = this.tokenizer.getAbsoluteIndex();
+
         if (this.lowerCaseTagNames) {
             name = name.toLowerCase();
         }
+
+        this.emitOpenTag(name);
+    }
+
+    private emitOpenTag(name: string) {
+        this.openTagStart = this.startIndex;
         this.tagname = name;
-        if (
-            !this.options.xmlMode &&
-            Object.prototype.hasOwnProperty.call(openImpliesClose, name)
-        ) {
-            let el;
+
+        const impliesClose =
+            !this.options.xmlMode && openImpliesClose.get(name);
+
+        if (impliesClose) {
             while (
                 this.stack.length > 0 &&
-                openImpliesClose[name].has(
-                    (el = this.stack[this.stack.length - 1])
-                )
+                impliesClose.has(this.stack[this.stack.length - 1])
             ) {
-                this.onclosetag(el);
+                const el = this.stack.pop()!;
+                this.cbs.onclosetag?.(el, true);
             }
         }
-        if (this.options.xmlMode || !voidElements.has(name)) {
+        if (!this.isVoidElement(name)) {
             this.stack.push(name);
             if (foreignContextElements.has(name)) {
                 this.foreignContext.push(true);
@@ -271,93 +285,117 @@ export class Parser {
         if (this.cbs.onopentag) this.attribs = {};
     }
 
-    onopentagend(): void {
-        this.updatePosition(1);
+    private endOpenTag(isImplied: boolean) {
+        this.startIndex = this.openTagStart;
+        this.endIndex = this.tokenizer.getAbsoluteIndex();
+
         if (this.attribs) {
-            this.cbs.onopentag?.(this.tagname, this.attribs);
+            this.cbs.onopentag?.(this.tagname, this.attribs, isImplied);
             this.attribs = null;
         }
-        if (
-            !this.options.xmlMode &&
-            this.cbs.onclosetag &&
-            voidElements.has(this.tagname)
-        ) {
-            this.cbs.onclosetag(this.tagname);
+        if (this.cbs.onclosetag && this.isVoidElement(this.tagname)) {
+            this.cbs.onclosetag(this.tagname, true);
         }
+
         this.tagname = "";
     }
 
+    /** @internal */
+    onopentagend(): void {
+        this.endOpenTag(false);
+
+        // Set `startIndex` for next node
+        this.startIndex = this.endIndex + 1;
+    }
+
+    /** @internal */
     onclosetag(name: string): void {
-        this.updatePosition(1);
+        this.endIndex = this.tokenizer.getAbsoluteIndex();
+
         if (this.lowerCaseTagNames) {
             name = name.toLowerCase();
         }
+
         if (
             foreignContextElements.has(name) ||
             htmlIntegrationElements.has(name)
         ) {
             this.foreignContext.pop();
         }
-        if (
-            this.stack.length &&
-            (this.options.xmlMode || !voidElements.has(name))
-        ) {
-            let pos = this.stack.lastIndexOf(name);
+
+        if (!this.isVoidElement(name)) {
+            const pos = this.stack.lastIndexOf(name);
             if (pos !== -1) {
                 if (this.cbs.onclosetag) {
-                    pos = this.stack.length - pos;
-                    while (pos--) {
+                    let count = this.stack.length - pos;
+                    while (count--) {
                         // We know the stack has sufficient elements.
-                        this.cbs.onclosetag(this.stack.pop() as string);
+                        this.cbs.onclosetag(this.stack.pop()!, count !== 0);
                     }
                 } else this.stack.length = pos;
-            } else if (name === "p" && !this.options.xmlMode) {
-                this.onopentagname(name);
-                this.closeCurrentTag();
+            } else if (!this.options.xmlMode && name === "p") {
+                this.emitOpenTag(name);
+                this.closeCurrentTag(true);
             }
-        } else if (!this.options.xmlMode && (name === "br" || name === "p")) {
-            this.onopentagname(name);
-            this.closeCurrentTag();
+        } else if (!this.options.xmlMode && name === "br") {
+            // We can't go through `emitOpenTag` here, as `br` would be implicitly closed.
+            this.cbs.onopentagname?.(name);
+            this.cbs.onopentag?.(name, {}, true);
+            this.cbs.onclosetag?.(name, false);
         }
+
+        // Set `startIndex` for next node
+        this.startIndex = this.endIndex + 1;
     }
 
+    /** @internal */
     onselfclosingtag(): void {
         if (
             this.options.xmlMode ||
             this.options.recognizeSelfClosing ||
             this.foreignContext[this.foreignContext.length - 1]
         ) {
-            this.closeCurrentTag();
+            this.closeCurrentTag(false);
+
+            // Set `startIndex` for next node
+            this.startIndex = this.endIndex + 1;
         } else {
+            // Ignore the fact that the tag is self-closing.
             this.onopentagend();
         }
     }
 
-    private closeCurrentTag() {
+    private closeCurrentTag(isOpenImplied: boolean) {
         const name = this.tagname;
-        this.onopentagend();
-        /*
-         * Self-closing tags will be on the top of the stack
-         * (cheaper check than in onclosetag)
-         */
+        this.endOpenTag(isOpenImplied);
+
+        // Self-closing tags will be on the top of the stack
         if (this.stack[this.stack.length - 1] === name) {
-            this.cbs.onclosetag?.(name);
+            // If the opening tag isn't implied, the closing tag has to be implied.
+            this.cbs.onclosetag?.(name, !isOpenImplied);
             this.stack.pop();
         }
     }
 
+    /** @internal */
     onattribname(name: string): void {
+        this.startIndex = this.tokenizer.getAbsoluteSectionStart();
+
         if (this.lowerCaseAttributeNames) {
             name = name.toLowerCase();
         }
         this.attribname = name;
     }
 
+    /** @internal */
     onattribdata(value: string): void {
         this.attribvalue += value;
     }
 
+    /** @internal */
     onattribend(quote: string | undefined | null): void {
+        this.endIndex = this.tokenizer.getAbsoluteIndex();
+
         this.cbs.onattribute?.(this.attribname, this.attribvalue, quote);
         if (
             this.attribs &&
@@ -380,47 +418,74 @@ export class Parser {
         return name;
     }
 
+    /** @internal */
     ondeclaration(value: string): void {
+        this.endIndex = this.tokenizer.getAbsoluteIndex();
+
         if (this.cbs.onprocessinginstruction) {
             const name = this.getInstructionName(value);
             this.cbs.onprocessinginstruction(`!${name}`, `!${value}`);
         }
+
+        // Set `startIndex` for next node
+        this.startIndex = this.endIndex + 1;
     }
 
+    /** @internal */
     onprocessinginstruction(value: string): void {
+        this.endIndex = this.tokenizer.getAbsoluteIndex();
+
         if (this.cbs.onprocessinginstruction) {
             const name = this.getInstructionName(value);
             this.cbs.onprocessinginstruction(`?${name}`, `?${value}`);
         }
+
+        // Set `startIndex` for next node
+        this.startIndex = this.endIndex + 1;
     }
 
+    /** @internal */
     oncomment(value: string): void {
-        this.updatePosition(4);
+        this.endIndex = this.tokenizer.getAbsoluteIndex();
+
         this.cbs.oncomment?.(value);
         this.cbs.oncommentend?.();
+
+        // Set `startIndex` for next node
+        this.startIndex = this.endIndex + 1;
     }
 
+    /** @internal */
     oncdata(value: string): void {
-        this.updatePosition(1);
+        this.endIndex = this.tokenizer.getAbsoluteIndex();
+
         if (this.options.xmlMode || this.options.recognizeCDATA) {
             this.cbs.oncdatastart?.();
             this.cbs.ontext?.(value);
             this.cbs.oncdataend?.();
         } else {
-            this.oncomment(`[CDATA[${value}]]`);
+            this.cbs.oncomment?.(`[CDATA[${value}]]`);
+            this.cbs.oncommentend?.();
         }
+
+        // Set `startIndex` for next node
+        this.startIndex = this.endIndex + 1;
     }
 
+    /** @internal */
     onerror(err: Error): void {
         this.cbs.onerror?.(err);
     }
 
+    /** @internal */
     onend(): void {
         if (this.cbs.onclosetag) {
+            // Set the end index for all remaining tags
+            this.endIndex = this.startIndex;
             for (
                 let i = this.stack.length;
                 i > 0;
-                this.cbs.onclosetag(this.stack[--i])
+                this.cbs.onclosetag(this.stack[--i], true)
             );
         }
         this.cbs.onend?.();
@@ -436,6 +501,8 @@ export class Parser {
         this.attribname = "";
         this.attribs = null;
         this.stack = [];
+        this.startIndex = 0;
+        this.endIndex = 0;
         this.cbs.onparserinit?.(this);
     }
 
